{*******************************************************}
{File:      NCOciUtil.PAS                               }
{Revision:  0.01.02 / 06.11.1999                        }
{Comment:   NC OCI8 VCL: utility unit                   }
{Copyright: (c) 1999, Nica-Com Ltd.                     }
{Author:    Dmitry Arefiev, diman@ncom.ru               }
{*******************************************************}
{$I NCOciDef.inc}

unit NCOciUtil;

interface

Uses NCOciDB;

function StrLike(const AStr, AMask: String; ANoCase: Boolean = False;
    AManyCharsMask: Char = '%'; AOneCharMask: Char = '_'; AEscapeChar: Char = '\'): Boolean;
function StrRPos(Str1, Str2: PChar): PChar;
function Var2SQL(const AValue: Variant; const ADataFormat: TOCIDataFormat): String;

implementation

Uses SysUtils;

function StrLike(const AStr, AMask: String; ANoCase: Boolean = False;
    AManyCharsMask: Char = '%'; AOneCharMask: Char = '_'; AEscapeChar: Char = '\'): Boolean;

    function InternalScan(pStr, pMask: PChar): Boolean;
    begin
        Result := True;
        while ((pStr^ = pMask^) and (pMask^ <> AManyCharsMask) or
               (pMask^ = AOneCharMask) or (pMask^ = AEscapeChar)) and
              (pStr^ <> #0) and (pMask^ <> #0) do begin
            if pMask^ = AEscapeChar then begin
                Inc(pMask);
                if pStr^ <> pMask^ then
                    Break;
            end;
            Inc(pMask);
            Inc(pStr);
        end;
        if pMask^ = AManyCharsMask then begin
            while (pMask^ = AManyCharsMask) or (pMask^ = AOneCharMask) do begin
                if pMask^ = AOneCharMask then
                    if pStr^ = #0 then begin
                        Result := False;
                        Exit;
                    end
                    else
                        Inc(pStr);
                Inc(pMask);
            end;
            if pMask^ <> #0 then begin
                while (pStr^ <> #0) and ((pMask^ <> pStr^) or not InternalScan(pStr, pMask)) do
                    Inc(pStr);
                Result := pStr^ <> #0;
            end;
        end
        else
            Result := (pMask^ = #0) and (pStr^ = #0);
    end;

begin
    if AMask = AManyCharsMask then
        Result := True
    else if (AMask = '') or (AStr = '') then
        Result := False
    else begin
        if ANoCase then
            Result := InternalScan(PChar(AnsiUpperCase(AStr)), PChar(AnsiUpperCase(AMask)))
        else
            Result := InternalScan(PChar(AStr), PChar(AMask));
    end;
end;

function StrRPos(Str1, Str2: PChar): PChar;
var
    p1, p2, pBSub, pESub, pBS, pES: PChar;
begin
    Result := nil;
    if (Str1 = nil) or (Str2 = nil) then
        Exit;
    pBSub := Str2;
    pESub := Str2 + StrLen(Str2) - 1;
    pBS := Str1 + StrLen(Str2) - 1;
    pES := Str1 + StrLen(Str1) - 1;
    while pES >= pBS do begin
        if pES^ = pESub^ then begin
            p1 := pES;
            p2 := pESub;
            while (p2 >= pBSub) and (p1^ = p2^) do begin
                Dec(p1);
                Dec(p2);
            end;
            if p2 < pBSub then begin
                Result := p1 + 1;
                Exit;
            end;
        end;
        Dec(pES);
    end;
end;

function Var2SQL(const AValue: Variant; const ADataFormat: TOCIDataFormat): String;
var
    V: Variant;
begin
    case VarType(AValue) and varTypeMask of
    varEmpty, varNull: Result := 'NULL';
    varSmallint, varInteger, varByte: Result := AValue;
    varSingle, varDouble, varCurrency: Result := FloatToStr(Extended(AValue));
    varDate: Result := 'to_date(''' + FormatDateTime('dd.mm.yyyy hh:nn:ss',
                       TDateTime(AValue)) + ''', ''dd.mm.yyyy hh24:mi:ss'')';
    varOleStr, varString: Result := '''' + AValue + '''';
    varBoolean:
        if ADataFormat <> nil then begin
            V := ADataFormat.Value2OBool(AValue);
            if (VarType(V) = varOleStr) or (VarType(V) = varString) then
                Result := '''' + V + ''''
            else
                Result := V;
        end
        else
            Result := '';
    else
        Result := '';
    end;
end;

end.


